// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Extract.h --- abstract class for product creation
//
//	DESCRIPTION:
//		This is a class definition for an abstract object Extractor
//		which is capable of creating product files (images, spectra,
//		rates file) from event list.
//
//	CHANGE HISTORY:
//		UL, 20/02/98: 1.0.0: - first draft
//              DAH, 25/03/98: 1.0.0: - first real version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//
//	AUTHOR:
//		D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description for SAS task `evselect'
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/evselect/src/Extractor.h,v 1.2 2008/06/09 16:33:23 aibarra Exp $
//

#ifndef EXTRACTOR
#define EXTRACTOR

#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>

class Extractor {
private:

public:
  // empty default constructor
  Extractor(){}  
 
  // empty destructor
  ~Extractor(){} 
 
  // do the product extraction
  void doIt(Table *, bool, const std::string &, int, bool);
};


class Product {

protected:

  Table * _intable;
  Table * _tempTable;
  bool _destruct;
  const std::string _flagCol;
  int32 _flagBit;

  DataSet *_dataSet;
  real64 _binSize;
  std::vector<std::string> _globalAttribs;

  void _copyGlobalAttribs(Block *out);
  void getRange(const Column *col, real64 &min, real64 &max);
  void GetLimits(const Column *, std::string minparam, std::string maxparam, bool useparam,
		 std::vector<real64> &limits);
  // AI: New method to calculate the Observation Start Time and Stop Time using the TSTART and TSTOP
  // SAS-SCR-238
  void GetObservationLimits(const Column *, std::string minparam, std::string maxparam, bool useparam,
		 std::vector<real64> &limits);
  Column * reType(const Column *oldCol, Dal::DataType newType);

public:

  // Product(){};

  Product(const std::string &name, Table *table, bool destructive,
	  const std::string &flagColumn, int flagBit);

  virtual void _getExposure(Block *productBlock);
  void _doColGen(ExprColumnGenerator &colGen, const Table &inTable, Table &outTable);

  // destructor
  virtual ~Product();

};

class Image : public Product {

private:
  vector<double> _radec2xy(vector<double> world, Column *Xcol, Column *Ycol);

public:

  Image(const std::string & name, Table *intable,
	bool destructive, const std::string &flagColumn, int flagBit) 
    : Product(name, intable, destructive, flagColumn, flagBit) {};

  void make(bool writeDss = false);

  void _getExposure(Block *productBlock);
}; 

class Spec : public Product {

public:

  Spec(const std::string & name, Table *intable,
	bool destructive, const std::string &flagColumn, int flagBit) 
    : Product(name, intable, destructive, flagColumn, flagBit) {};

  // Check that the spectrum has standard selections and ranges
  void checkStandard(std::vector<int32> piRange); 

  void make(bool writeDss = false);
}; 

class Rate : public Product {

private:

public:

  Rate(const std::string & name, Table *intable,
	bool destructive, const std::string &flagColumn, int flagBit) 
    : Product(name, intable, destructive, flagColumn, flagBit) {};
  void make(bool writeDss = false);
}; 

class Histo : public Product {

private:

public:

  Histo(const std::string & name, Table *intable,
	bool destructive, const std::string &flagColumn, int flagBit) 
    : Product(name, intable, destructive, flagColumn, flagBit) {};
  void make(bool writeDss = false);
}; 

#endif

