
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//  evselect.cc --- event selection task
//
//  DESCRIPTION:
//      evselect is used to filter event list with a boolean expression
//      and construct images, spectra, and rates file from the new list.
//
//  CHANGE HISTORY:
//      UL, 1998-01-06: 0.0.1: - skeleton
//      UL, 1998-03-20: 1.0.0: - first useful version
//      UL, 1998-04-02: 1.1.0: - extended functionaliy (new region filtering
//                               functions)
//      UL, 1998-04-06: 1.2.0: - new region filters (in line with ASC region
//                               filters); destructive filtering mode works
//                               now (fix of bug in DAL)
//      UL, 1998-04-20: 1.3.0: - check for existence of input/output event
//                               list and consistency with clobber value
//                             - selectcore v1.1.0 (speed-impoved GTI
//                               filtering)
//      DH, 1998-??/??: 1.4.0: - ???
//      UL, 1998-04-27: 1.5.0: - add parameter values to file history
//                             - hard-coded version number removed (fix SPR
//                               170)
//      UL, 1998-05-28: 1.6.0: - close SSC-SPR 169 about generic main
//      UL, 1998-06-29: 1.7.0: - brought in line with selectlib v3.0
//      UL, 1998-11-05: 1.8.0: - open input table in DataSet::Read mode
//                               if it is not to be modified
//                               (filterpersist=false); this is to speed
//                               up the close() call at the end which
//                               in DataSet::Modify mode always writes
//                               back the data set to disk
//                             - got rid of uneccessary #include statements
//                             - fixed bug in extraction of block specifier
//      UL, 1999-02-11: 1.8.1: - added include statement enforced by
//                               taskmain >v1.17
//      UL, 1999-04-16: 1.8.2: - brought it in-line with selectlib-4.0
//                               (Filter -> DataFilter)
//      UL, 1999-05-11: 1.8.3: - brought in line with error-1.18
//      UL, 1999-05-12: 1.9.0: - made use of new method in
//                               selectlib::DataFilter class to replace
//                               symbolic references to table attributes
//                               by their values in user-specified selection
//                               expression; this enables makes it possible
//                               to have entire subexpressions in string
//                               attributes and let the user make use of
//                               it from the command line, e.g.
//                               in the table
//                                  SEL_DIAG = 'FLAG & 0xff12 != 0' / bla bla
//                               ...
//                               unix> evselect table=ev.fits
//                                          expression='#SEL_DIAG' ...
//		DH, 2000-01-31: 2.0.0: - support for DSS
//							   - overhauled file access modes
//		UL, 2000-02-04: 2.1.0: - overhauled code
//							   - for keepfilteredoutput=no the input
//								 data set is now copied to a Temp
//								 (Dal terminology) before it is being
//								 destructively filterd; in the highmem case
//								 this has the advantage that the modified
//								 table is never written back to disk
//							   - experimental parameter "tempsetmemenv"
//
//  AUTHORS:
//      U. Lammers, Tue Jan  6 17:13:17 MET 1998
//		D. Hinshaw
//
//  REFERENCES:
//      [1]: Software Specification Document for the XMM Science Analysis
//           Subsystem, M. Dahlem et. al.
//      [2]: Evselect task description, D. Hinshaw (SSC), U. Lammers (SOC)
//
//  $Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/evselect/src/evselect.cc,v 1.5 2009/05/11 10:43:04 aibarra Exp $
//

#include <cstdio>

#include <fstream>
#include <string>

#include "param.h"
#include "errstr.h"
#include "Dal.h"
#include "dal_utilities.h"
#include "FileLocator.h"
#include "stringutil.h"
#include "DataFilter.h"
#include "Extractor.h"
#include "SelectUtils.h"
#include "DataSubspace.h"
#include "DssBuilder.h"
#include "xmmexposure.h"

#include "evselect.h"

void evselect()
{
	PROC

	//
	//	Determine if the expression constitutes a 'do nothing' filter
	//
	std::string expression;
	bool nullFilter = false;
	if(stringParameter("filtertype") == "expression" ) {
		expression = stringParameter("expression");
		nullFilter = expression == "" ||
				 expression == "-" ||
				 uppercase(expression) == "TRUE";
	}

	//  Get the input set and mode based on the values of withfilteredset
	//	and keepfilteroutput
	bool withFilteredSet	= booleanParameter("withfilteredset"),
	keepFilteredOutput		= booleanParameter("keepfilteroutput");

	if( !keepFilteredOutput && !booleanParameter("withimageset") &&
            !booleanParameter("withspectrumset") && !booleanParameter("withrateset") && !booleanParameter("withhistogramset") ){
	  errstr << "You have requested that no output be produced.  Exiting." << Sas::warning("NoOutput");
	  return;
	}

	std::vector<std::string> v(breakup(stringParameter("table"), ":"));
	if (!v.size())
		errstr<<"Specified table name is empty"<<Sas::error("EmptyTableName");

	std::string inSetName = v[0],
		   inTableName = v.size() > 1 ? v[1] : "";

	DataSet *set;
	Table	*table;
	if (keepFilteredOutput)
		if (withFilteredSet) {
			//
			//	want to keep result of filtering stage and have it written
			//	to a new data set: copy the input set and open the copy
			//	in Modify mode if result of filtering is not NULL
			//
			std::string filteredSetName = stringParameter("filteredset");

			if(nullFilter){
			  errstr<<"Copying data ... "<<Sas::message(Msg::App, Msg::Verbose);
			  dataSetServer->copy(inSetName, filteredSetName);
			  errstr<<"done."<<Sas::message(Msg::App,Msg::Verbose);
			  set = dataSetServer->open(filteredSetName, Dal::Read);
			}else{
			  errstr<<"Cloning dataset ... "<<Sas::message(Msg::App, Msg::Verbose);
			  set = dataSetServer->clone(inSetName, filteredSetName, Dal::Modify);
			  errstr<<"done."<<Sas::message(Msg::App,Msg::Verbose);
			}
		} else {
			//
			//	want to keep the result of filtering stage by modifiying
			//	the input set - open input set in Modify mode unless
			//	the result of the filtering is NULL
			//
			set = dataSetServer->open(inSetName, !nullFilter ? Dal::Modify : Dal::Read);
		}
	else
	  //
	  //	do not want to keep the output of filtering stage:
	  //	make a copy to a Temp data set and open this in Modify
	  //	mode or open input set in Read mode if filter has no effect
	  //
	  if (nullFilter)
	    set = dataSetServer->open(inSetName, Dal::Read);
	  else
	    set = dataSetServer->clone(inSetName, tempnam(0,0), Dal::Temp);

	table = inTableName.length() ? set->table(inTableName) : set->table(0);
	bool noInputEvents = table->rows()==0;
	if(noInputEvents)
	  errstr<<"Input table contains no events." <<Sas::warning("NoInputEvents");

	//
	// Choose between filtering with a selection expression or a data subspace.
	//
	if (stringParameter("filtertype") == "dataSubspace") {
		std::string dssSpec = stringParameter("dssblock");
		DataSet *dssSet = dataSetServer->open(dssSpec, DataSet::Read);
		DataSubspace dss(dssSet->block(dssSpec));
		expression = dss.selectExpression();
		dataSetServer->close(dssSet);
	} else {
            // If the selection expression contains RA/DEC coords need to 
            // convert to XY
                //expression = convExpressionCeltoSky(expression, *table);
            // convert the IMAGE selection to BOX + MASK
            // and convert RA,DEC to X/Y
                if (!nullFilter) {
                   XmmExposure xmmexp2(table);
                   expression =  xmmexp2.convSpecialExpression(expression, inSetName);
                }

		expression = DataFilter::expandSelectionExpression(expression, *table);
        }
	//
	//  instantiate filter and do the filtering
	//
	bool destructive = booleanParameter("destruct");
	std::string flagColumn = stringParameter("flagcolumn");
	int flagBit = intParameter("flagbit");

	if (!nullFilter ) {
		DataFilter filter(expression);
		errstr<<"Filtering ... "<<Sas::message(Msg::App,Msg::Verbose);

		unsigned long selectedEvents =
				destructive	? filter.doFilter(*table) // destructive filtering
						: filter.doSelect(*table, flagColumn, flagBit); // non-destructive filtering

		errstr << "selected " << selectedEvents << " rows from the input table." << Sas::message(Msg::App,Msg::Sparse);

		if (!selectedEvents && !noInputEvents) {
			errstr<<"No events have been selected";
			if (destructive)
			  errstr<<" - filtering of the event list resulted in an empty table";
			errstr<<Sas::warning("NoFilteredEvents");
		}
		errstr<<"done"<<Sas::message(Msg::App,Msg::Verbose);

		//
		//	Construct the data subspace for this filtering expression, if
		//	requested.
		//
		bool dssError = false;
		if( booleanParameter("writedss") ){
		        DataSubspace dss, newDss;
			newDss = dssBuilder.getDss(expression, table);
			std::string checkType = "DSTYP1";
			if( table->hasAttribute(checkType) ){
			  table >> dss;
			  deleteDSS(table);
			  dss *= newDss;
			  //Creates the GTI extensions
			  dss.compressRegions();
			}else
			  dss = newDss;

			if( booleanParameter("cleandss") ) cleanDSS(table, dss);

			table<<dss;
			dssError = newDss.error();
		}

		*table->addAttribute("SLCTEXPR", "Filtering expression used by "
				     "evselect") = expression;
		//
		// Update exposure information, if requested
		//
		if( booleanParameter("updateexposure") ){
		  if( !booleanParameter("writedss") )
		    errstr << "Exposure update requested without writing of the data subspace. "
			   << "Exposure information is likely to be incorrect." << Sas::warning("ExpNoDss");

		  if( dssError )
		    errstr << "Exposure update requested when there was a data subspace error. "
			   << "Exposure information is likely to be incorrect." << Sas::warning("ExpBadDss");

                  // 
		  // Only need to do exposure update if TIME filtering was done
		  // AI: SPR-2632. Now also we check if there is a PHASE filtering. In this case
		  // we also have to update the LIVETIME keyword.
		  // We also check if the FRMTIME is bigger than PERIOD (phasecalc now write the
		  // PERIOD kwyword to the header. If this is the case, we raise a warning message

		  XmmExposure xmmexp(table);
		  if( (expression.find("TIME") != std::string::npos) ||
		      (expression.find("PHASE") != std::string::npos) )
		    {
		      if ( (expression.find("PHASE") != std::string::npos) )
		      {
		    	  double frmtime = 0.0;
		    	  if (table->hasAttribute("FRMTIME"))
		    		  frmtime = table->attribute("FRMTIME")->asReal()/1000.0;
		    	  else
		    	  {
		    		  //Seek for FRMTIME keyword
		    		  for(unsigned i = 0; i < set->blocks(); i++){
		    		    if( set->table(i)->hasAttribute("FRMTIME"))
		    		    {
		    		    	frmtime = set->table(i)->attribute("FRMTIME")->asReal()/1000.0;
		    		    	break;
		    		    }
		    		  }
		    		  if (frmtime == 0.0)
		    			  errstr << "No FRMTIME keyword found while cheking PHASE expression "
		    			  <<  Sas::warning("NoFrameTimeKeywordFound");

		    	  }

		    	  if (table->attribute("PERIOD")->asReal() <= frmtime )
		    	  {
		    		  errstr << "The Frame Time value for this mode is bigger than the period. "
		    		  << "Frame Time value: "
		    		  << frmtime/1000.0 << " secs. "
		    		  << "Period: " <<table->attribute("PERIOD")->asReal()  << " secs. "
		    		  << Sas::warning("FrameTimeBiggerThanPeriod");
		    	  }
		      }
		      xmmexp.getExposureInformation();
		      xmmexp.updateOntime();
		      xmmexp.updateLivetime();
		    } // End of if time or phase selection block

                  // Find the number of discarded lines measured during the selected times
                  // and write as header keywords if EPIC-pn
                  if (table->hasAttribute("INSTRUME")) {

                    // Xmm::toInstrument produces error if 
                    // instrument not recognised so use string
                     string inst = table->attribute(
                                "INSTRUME")->asString();
                     if (uppercase(inst).find("EPN") 
                                         != string::npos) {
                        xmmexp.writeDSlineKeys();
                     }
                  }
		} // End of if updateexposure block
	}

	if( keepFilteredOutput ){
		// Set HDUCLAS2 keyword
		if (table->hasAttribute("HDUCLAS1") &&
			table->attribute("HDUCLAS1")->asString() == "EVENTS") {
				if (!table->hasAttribute("HDUCLAS2"))
					table->addAttribute("HDUCLAS2");
				*table->attribute("HDUCLAS2") = destructive ? "ACCEPTED" : "ALL";
		}
	}

	//
	//  extract products
	//
	errstr<<"Product extraction ..."<<Sas::message(Msg::App,Msg::Verbose);
	Extractor extractor;
	extractor.doIt(table, (destructive || nullFilter), flagColumn, flagBit,
				   booleanParameter("writedss"));
	errstr<<"done"<<Sas::message(Msg::App,Msg::Verbose);

	//  close data set
	errstr<<"closing data set ... "<<Sas::message(Msg::App,Msg::Verbose);
	dataSetServer->close(set);
	errstr<<"done"<<Sas::message(Msg::App,Msg::Verbose);

	ENDPROC
} // evselect()

