// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	spec.cc -- implementation of the Spec class
//
//	CHANGE HISTORY:
//              DAH, 25/03/98: 1.0.0: - first version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//				UL, 2000-12-20: 2.0.0: - couple of trivial changes to make the
//										 optimizer happy
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/evselect/src/spec.cc,v 1.2 2004/07/09 08:56:19 aibarra Exp $

#include <iostream>
#include <sstream>
#include <cmath>
#include <cstdio>

#include "param.h"
#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "Extractor.h"
#include "DataSubspace.h"
#include "DataFilter.h"
#include "xmmexposure.h"

void Spec::make(bool writeDss){

  // Read in the needed parameters

  int IbinSize = intParameter("spectralbinsize");
  bool zcol = booleanParameter("withzcolumn");
  bool zerrcol = booleanParameter("withzerrorcolumn");

  // Get pointers to the needed columns

  Column *eDatCol, *zDatCol = 0, *zDatErrCol = 0;
  Column *eCol, *zCol, *zErrCol;

  std::string ecolumn = stringParameter("energycolumn");
  if( _intable->hasColumn(ecolumn) ){
    eCol = _intable->column(ecolumn);

    if( eCol->dataType()!=Dal::Int8 && eCol->dataType()!=Dal::Int16 && 
	eCol->dataType()!=Dal::Int32 )
      errstr << "Energy column is not integer-valued.  All values will be "
             << "truncated to integers before accumulation of spectrum."
	     << Sas::warning("SpecIntColumn");

    eDatCol = reType(eCol, Dal::Int32);
  }else{
    std::stringstream newName;
    newName << "E" << _intable;
    ExprColumnGenerator colGen(ecolumn, newName.str(), Dal::Int32);
    _doColGen(colGen, *_intable, *_tempTable);
    eCol = eDatCol = _intable->column(ecolumn);
  }

  if(zcol){
    if( !_intable->parent()->hasAttribute("TELESCOP") ||
	_intable->parent()->attribute("TELESCOP")->asString() != "XMM" ){
      errstr << "Not XMM data.  Cannot calculate a weighted spectrum for non-XMM data"
	     << Sas::warning("NotXmmSpectrum");
      return;
    }

    std::string zcolumn = stringParameter("zcolumn");
    if( _intable->hasColumn(zcolumn) ){
      zCol = _intable->column(zcolumn);
      zDatCol=reType(zCol, Dal::Real64);
    }else{
      std::stringstream newName;
      newName << "Z" << _intable;
      ExprColumnGenerator colGenZ(zcolumn, newName.str());
      _doColGen(colGenZ, *_intable, *_tempTable);
      zCol = zDatCol = _intable->column(zcolumn);
    }

    if(zerrcol){
      zcolumn = stringParameter("zerrorcolumn");
      if( _intable->hasColumn(zcolumn) ){
	zErrCol = _intable->column(zcolumn);
	zDatErrCol=reType(zErrCol, Dal::Real64);
      }else{
	std::stringstream newName;
	newName << "ZE" << _intable;
	ExprColumnGenerator colGenZ(zcolumn, newName.str());
	_doColGen(colGenZ, *_intable, *_tempTable);
	zErrCol = zDatErrCol = _intable->column(zcolumn);
      }
    }
  }

  // Determine the energy range

  std::vector<real64> limits(2);
  GetLimits(eCol, "specchannelmin", "specchannelmax", booleanParameter("withspecranges"), limits);

  std::vector<int32> useRange(2);
  useRange[0] = (int32)limits[0];
  useRange[1] = (int32)limits[1];
  std::vector<int32> tlRange(2);

  if( useRange[1] == useRange[0] )
    errstr << "The energy column for extracting the spectrum has "
	   << "the same value throughout, or the upper and lower limits "
	   << "for this column are equal. An empty spectrum will "
	   << "be created." << Sas::warning("BadSpecLimits"); 

  if( useRange[1] < useRange[0] )
    errstr << "Upper spectral limit < lower limit.  Cannot continue." 
	   << Sas::error("BadSpecLimits");

  unsigned long nRows = (unsigned long) ceil( ((float)(useRange[1]-useRange[0]+1))/IbinSize );

  // Check if a standard spectrum has been requested
  checkStandard(useRange);

  // Now initialize the output spectrum

  Table * outTable = _dataSet->addTable("SPECTRUM", nRows);

  _copyGlobalAttribs(outTable);

  // Write some keywords

  *outTable->addAttribute( "HDUCLASS",
			   "Format conforms to OGIP/GSFC conventions")
    = "OGIP";
  *outTable->addAttribute( "HDUCLAS1",
			    "File contains a spectrum")
    = "SPECTRUM";
    
  if (zcol)
    {
      *outTable->addAttribute( "HDUCLAS2",
			       "File contains gross rates")
	= "TOTAL";
      
      *outTable->addAttribute( "HDUCLAS3",
			       "Spectrum is stored as rates")
	= "RATE";
    }
  else
    {
      *outTable->addAttribute( "HDUCLAS2",
			       "File contains gross counts")
	= "TOTAL";
      
      *outTable->addAttribute( "HDUCLAS3",
			       "Spectrum is stored as counts")
	= "COUNT";
    }
  *outTable->addAttribute( "HDUVERS1",
			   "Version of format")
    = "1.1.0";

  *outTable->addAttribute( "AREASCAL",
			   "Nominal scaling factor for data")
    = 1.0;

  *outTable->addAttribute( "BACKSCAL",
			   "Scaling factor for background")
    = 1.0;

  *outTable->addAttribute( "CORRSCAL",
			   "Nominal scaling factor for correction file")
    = 1.0;

  if (zcol)
    {
      *outTable->addAttribute( "POISSERR",
			       "Poisson errors appropriate")
	= false;
    }
  else
    {
      *outTable->addAttribute( "POISSERR",
			       "Poisson errors appropriate")
	= true;
    }

  *outTable->addAttribute( "SYS_ERR",
			   "Global systematic error")
    = 0;

  *outTable->addAttribute( "CHANTYPE","Type of channel data") = eCol->name();

  *outTable->addAttribute( "DETCHANS",
			   "Total number of detector channels available")
    = (int)nRows;

  *outTable->addAttribute( "QUALITY",
			   "All channels have good quality")
    = 0;

  *outTable->addAttribute( "GROUPING",
			   "No data grouping done")
    = 0;

  *outTable->addAttribute( "SPECDELT",
			   "Spectral channel size, ie the binning factor")
    = IbinSize;

  *outTable->addAttribute( "SPECPIX",
			   "The rebinned channel correspondsing to SPECVAL")
    = 0;

  *outTable->addAttribute( "SPECVAL",
			   "Original chan value at center of rebinned chan")
    = useRange[0] - 0.5 + float(IbinSize)/2;

  // Set related file keywords to NONE
  *outTable->addAttribute( "BACKFILE",
                           "Name of background file") = "NONE";
  *outTable->addAttribute( "ANCRFILE",
                           "Name of ARF file") = "NONE";
  *outTable->addAttribute( "RESPFILE",
                           "Name of response file") = "NONE";

  // Make the necessary columns.
  
  Column * chanCol = outTable->addColumn("CHANNEL", Dal::Int16, "");
  ColumnData * cChanDat = chanCol->data();
  int16 * chanDat = cChanDat->int16Data();

  *chanCol->addAttribute("TLMIN") = 0;

  *chanCol->addAttribute("TLMAX") = (int)nRows-1;

  real64 *rateDat=0, *rateErrDat=0;
  real64 *zData=0, *zErrData=0;
  ColumnData *cRateDat=0, *cErrRateDat=0, *cZDatCol=0, *cZDatErrCol=0, *cCntDat=0;
  int32 * cntDat=0;

  if( zcol ){
    cRateDat = outTable->addColumn("RATE", Dal::Real64, "", "count/s")->data();
    cErrRateDat = outTable->addColumn("STAT_ERR", Dal::Real64, "", "count/s")->data();
    rateDat = cRateDat->real64Data();
    rateErrDat = cErrRateDat->real64Data();
    for(unsigned i=0; i<nRows; i++) rateDat[i] = rateErrDat[i] = 0;

    cZDatCol = zDatCol->data();
    zData = cZDatCol->real64Data();
    if(zerrcol){
      cZDatErrCol = zDatErrCol->data();
      zErrData = cZDatErrCol->real64Data();
    }
  }else{
    cCntDat = outTable->addColumn("COUNTS", Dal::Int32, "", "count")->data();
    cntDat = cCntDat->int32Data();
    for(unsigned i=0; i<nRows; i++) cntDat[i]=0;
  }
  
  for(unsigned i=0; i<nRows; i++)
    chanDat[i] = i;

  delete cChanDat;

  // Set up type converted columns

  Column *flagDatCol;
  ColumnData *cFlagData = 0, *cEData = eDatCol->data();
  int32 index, *flagData = 0, *eData = cEData->int32Data();

  if(! _destruct){
    flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }

  // Accumlate the spectrum

  unsigned long mRows = _intable->rows();
 
  if(zcol){
    // Get the CCD info necessary for calculating rates
    std::vector<real64> exposures(18);
    for(int i=1; i<=17; i++){
      std::stringstream attName;
      attName << "LIVETI";
      attName.width(2);
      attName.fill('0');
      attName << i;
      if(_intable->hasAttribute(attName.str()))
	exposures[i] = _intable->attribute(attName.str())->asReal();
    }

    ColumnData * cCcdDat = _intable->column("CCDNR")->data();
    int8 * ccdDat = cCcdDat->int8Data();
    
    for(unsigned long irow=0; irow<mRows; irow++){
      if( ! _destruct && (( flagData[irow] & _flagBit) == 0) ) continue;

      index = (eData[irow]-useRange[0])/IbinSize;
      if( index >= 0 && index < (int32)nRows ){
	rateDat[index] += zData[irow]/exposures[ccdDat[irow]];
	if(zerrcol)
	  rateErrDat[index] += zErrData[irow]*zErrData[irow]/(exposures[ccdDat[irow]]*exposures[ccdDat[irow]]);
	else
	  rateErrDat[index] += zData[irow]*zData[irow]/(exposures[ccdDat[irow]]*exposures[ccdDat[irow]]);
      }
    }

    for(unsigned long irow=0; irow<outTable->rows(); irow++)
      rateErrDat[irow] = sqrt(rateErrDat[irow]);
    
    delete cRateDat;
    delete cErrRateDat;
    delete cZDatCol;
    if(zerrcol) delete cZDatErrCol;

  }else{
    for(unsigned long irow=0; irow<mRows; irow++){
      if( ! _destruct && (( flagData[irow] & _flagBit) == 0) ) continue;
      index = (eData[irow]-useRange[0])/IbinSize;
      if( index >= 0 && index < (int32)nRows ) cntDat[index]++;
    }
    delete cCntDat;
  }

  delete cEData;

  if(! _destruct) delete cFlagData;

  // Write data subspace info
  if (writeDss) {
     DataSubspace dss(_intable);
     outTable << dss;
  }

  // Calculate exposure keyword

  _getExposure(outTable);

  // Done
}

// RDS - check that a standard spectrum has been requested
void Spec::checkStandard(std::vector<int32> piRange) {

  // Check that the spectral range and pattern selections are as expected
  // and can be handled by later software, unless an override parameter
  // has been set. Only makes sense if instrument is known
  if (! booleanParameter("nonStandardSpec") && 
           _intable->parent()->hasAttribute("INSTRUME"))
   {
     // Check the PI channels selected for the spectrum which should be:
     // PN: 0-20479
     // MOS: 0-11999
     //
     // Check the pattern range which should be:
     // PN: singles, doubles or singles+doubles
     // MOS: singles or s+d+t+q
     //
     DataSubspace dss(_intable);
     list <Range<double> > ranges = dss.ranges("PATTERN");
     double pattlow=12, patthigh=0;
     double upper, lower;
     RangeBound upperB, lowerB;
     unsigned toohigh = 99; // A large number 

     for ( list<Range<double> >::iterator i = ranges.begin();
             i != ranges.end(); ++i )
       {
         (*i).values(lower, upper, lowerB, upperB);

         if (lowerB==Unbounded)
            lower = 0;
         else if (lowerB==Open)
            ++lower;

         if (upperB==Unbounded)
            upper  = toohigh;
         else if (upperB==Open)
            --upper;

         pattlow = lower < pattlow ? lower:pattlow;
         patthigh = upper > patthigh ? upper:patthigh;
       }
     // Make the check depending on instrument
     std::string instrument = _intable->parent()->
                                        attribute("INSTRUME")->asString();
     if (instrument=="EMOS1" || instrument=="EMOS2")
       {
        if (pattlow != 0 or (patthigh != 0 and patthigh != 12))
           errstr << "The standard event PATTERN range for a MOS "
           << "spectrum is 0-0 or 0-12. "
           << "Detector matrices can not be created for this spectrum "
           << "and spectral fitting will not be possible."
           << Sas::warning("NonStandardSpectrum");

        if (piRange[0] != 0 or piRange[1] != 11999)
           errstr << "The standard PI channel range for MOS is 0-11999. "
           << "Detector matrices can not be created for this spectrum "
           << "and spectral fitting will not be possible."
           << Sas::warning("NonStandardSpectrum");
       }
     else if (instrument=="EPN")
       {
        if ( (pattlow>1) or (pattlow==0 and !(patthigh==0 or patthigh==4))
                        or (pattlow==1 and patthigh!=4)  )
           errstr << "The standard event PATTERN range for a PN "
           <<"spectrum is 0-0, 0-1, 0-4 or 1-4. "
           << "Detector matrices can not be created for this spectrum "
           << "and spectral fitting will not be possible."
           << Sas::warning("NonStandardSpectrum");

        if (piRange[0] != 0 or piRange[1] != 20479)
           errstr << "The standard PI channel range for PN is 0-20479. "
           << "Detector matrices can not be created for this spectrum "
           << "and spectral fitting will not be possible."
           << Sas::warning("NonStandardSpectrum");
       }
    }
}

