// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//	xmmexposure.cc --- implementation of the XmmExposure class
//
//	DESCRIPTION:
//
//	AUTHOR:
//		D. Hinshaw
//

#include <cstdio>
#include <unistd.h>
#include <iostream>
#include <sstream>
#include <cctype>
#include <numeric>

#include "Xmm.h"

#include "param.h"
#include "errstr.h"
#include "Dal.h"
#include "DataSubspace.h"
#include "DataFilter.h"
#include "DssBuilder.h"
//#include "ServiceUtils.h"
#include "xmmexposure.h"
#include "OGIPgti.h"
#include "SscWcsConvertor.h"
#include "CoordPair.h"

XmmExposure::XmmExposure(Table *eventTable) : _eventTable(eventTable) {

  _set = _eventTable->parent();
  _instrument = Xmm::INSTRUMENT_NONE;

  // Do some checks to make sure we can do exposure correction on this data set

  if( !_set->hasAttribute("TELESCOP") ||
      _set->attribute("TELESCOP")->asString() != "XMM" ){
    errstr << "Can only update exposure information on XMM data." << Sas::warning("NotXMM");
    return;
  }

  _instrument = Xmm::toInstrument(_set->attribute("INSTRUME")->asString());

  // Get the list of CCDs with exposure extensions

  if( !_eventTable->hasColumn("CCDNR") ){
      errstr << "Event list does not have a CCDNR column. "
	     << "Cannot update exposure information. "
	     << Sas::warning("NoCCDcolumn");
      return;    
  }

  Column *ccdnr = _eventTable->column("CCDNR");
  ColumnData * cCcdnrDat = ccdnr->data();
  int8 * ccdnrDat = cCcdnrDat->int8Data();

  for(int i=1; i<=17; i++){
    std::stringstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << i;
    if( !_set->hasBlock(extName.str()) ) continue;

    bool hasCcd=false;
    for(unsigned irow=0; irow<_eventTable->rows(); irow++){
      if(i==ccdnrDat[irow]){
	hasCcd=true;
	break;
      }
    }

    if(hasCcd)
      _ccdNums.push_back(i);
    else
      _set->deleteBlock(extName.str());

    
    /*
     * AI: Check if there is a PHASE keyword in the expression. If so, check that 
     * the user has run phasecalc for the EXPOSU table, if not raise a warning message.
     */
    DataSubspace dss(_eventTable);
    unsigned ncomps = dss.components();
    for(unsigned icomp=0; icomp<ncomps; icomp++)
      {
	DScomp * comp = dss.component(icomp);
	if( comp->hasFilter("PHASE", DSS::RangeT) )
	  { 
	    if( _set->hasBlock(extName.str()) )	    
	      if (!_set->table(extName.str())->hasColumn("PHASE") )
		errstr << "You have used the PHASE keyword to filter the event list.  "
		       << "You have to run phasecalc task for all the EXPOSUnn tables"
		       << Sas::error("PhaseTimeFilter");
	  }
      }
  }

  return;
}


void XmmExposure::getExposureInformation() {

  if( _instrument==Xmm::EMOS1 || _instrument==Xmm::EMOS2 ||
      _instrument==Xmm::EPN ){
    if(_instrument==Xmm::EPN) _targetCCD = 4;
    else _targetCCD = 1;
    _getGtiLists();
  }else if( _instrument==Xmm::RGS1 || _instrument==Xmm::RGS2 ){

    // Check for any time based filters

    DataSubspace dss(_eventTable);
    unsigned ncomps = dss.components();

    for(unsigned icomp=0; icomp<ncomps; icomp++){
      DScomp * comp = dss.component(icomp);
      if( comp->hasFilter("TIME", DSS::RangeT) || comp->hasFilter("TIME", DSS::RangeBlockT) ||
	  comp->hasFilter("TIME", DSS::GtiT) ){
	errstr << "You have used the TIME column to filter an RGS events list.  "
	       << "This will invalidate any exposure information in the produced products."
	       << Sas::warning("RgsTimeFilter");
	return;
      }
    }
  }
  
  return;
}


void XmmExposure::updateOntime() {

  if(_instrument==Xmm::INSTRUMENT_NONE ||
	 _instrument==Xmm::RGS1 ||
	 _instrument==Xmm::RGS2) return;

  // Sum up all the GTIs to get the ontime    

  for(unsigned iccd=0; iccd<_ccdNums.size(); iccd++){
    double ontime = 0;
    assert(_gtiLists.size() > iccd);
    for(unsigned igti=0; igti<_gtiLists[iccd].size(); igti++)
      ontime += _gtiLists[iccd][igti].stop() - _gtiLists[iccd][igti].start();

    // Write the new ontime keywords.

    std::stringstream keyName;
    keyName << "ONTIME";
    keyName.width(2);
    keyName.fill('0');
    keyName << _ccdNums[iccd];
    
    *_eventTable->addAttribute(keyName.str(), "Sum of GTIs for a particular CCD") 
      = ontime;
  }

  // Write ONTIME keyword for target CCD

  std::stringstream keyName;
  keyName << "ONTIME";
  keyName.width(2);
  keyName.fill('0');
  keyName << _targetCCD;

  if( _eventTable->hasAttribute(keyName.str()) )
    *_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") 
      = _eventTable->attribute(keyName.str())->asReal();
  else if( _ccdNums.size()>0 ){
    errstr << "Ontime for target "
	   << "CCD (number " << _targetCCD << ") is not available. "
	   << "Setting ONTIME to zero."
	   << Sas::warning("NoTargetOntime");
    *_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") = 0;
  }

	// AI: SOC-SPR-2629:
	// If there are not events as a result of the filtering, the _ccdNums vector 
	// (the vector with the ccd numbers that contains events) is empty and therefore
	// the _gtiList is also empty, so the updateOntime method does not uptdate the ONTIME
	// keyword. 
	// To fix this problem, if the _ccdNums vector is empty, we check for all the 
	// ONTIME keywords in the header file and they are updated with 0 time. 
	if (_ccdNums.size() == 0)
	{
		errstr << "No events have been selected. Setting ONTIME values to 0..."
			   << Sas::message(Msg::App,Msg::Verbose);	
		double ontime = 0; 
		for(unsigned i=0; i<17; i++)
		{
			std::stringstream keyName;
  			keyName << "ONTIME";
  			keyName.width(2);
  			keyName.fill('0');
  			keyName << i;
  
			if( _eventTable->hasAttribute(keyName.str()) )			   			   		
	    		*_eventTable->addAttribute(keyName.str(), "Sum of GTIs for a particular CCD") 
    	  		= ontime;	
  		}
		if( _eventTable->hasAttribute("ONTIME") )			   			   		
	    	*_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") = ontime;	
	}

  return;
}


void XmmExposure::updateLivetime(){

  if(_instrument==Xmm::INSTRUMENT_NONE ||
	 _instrument==Xmm::RGS1 ||
	 _instrument==Xmm::RGS2) return;

  for(unsigned i=0; i<_ccdNums.size(); i++){
    std::stringstream keyName;
    keyName << "LIVETI";
    keyName.width(2);
    keyName.fill('0');
    keyName << _ccdNums[i];
    
    *_eventTable->addAttribute(keyName.str(), "Live time for a particular CCD") = _getLiveTime(i);
  }

  // Write LIVETIME keyword for the central CCD

  std::stringstream keyName;
  keyName << "LIVETI";
  keyName.width(2);
  keyName.fill('0');
  keyName << _targetCCD;
  
  if( _eventTable->hasAttribute(keyName.str()) )
    *_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") 
      = _eventTable->attribute(keyName.str())->asReal();
  else if( _ccdNums.size()>0 ){
    errstr << "Live time for target "
	   << "CCD (number " << _targetCCD << ") is not available. "
	   << "Setting LIVETIME to zero."
	   << Sas::warning("NoTargetLivetime");
    *_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") = 0;
  }


	// AI: SOC-SPR-2629:
	// If there are not events as a result of the filtering, the _ccdNums vector 
	// (the vector with the ccd numbers that contains events) is empty and therefore
	// the _gtiList is also empty, so the updateOntime method does not uptdate the LIVETIME
	// keyword. 
	// To fix this problem, if the _ccdNums vector is empty, we check for all the 
	// LIVETIME keywords in the header file and they are updated with 0 time. 
	if (_ccdNums.size() == 0)
	{		
		errstr << "No events have been selected. Setting LIVETIME values to 0..."
			   << Sas::message(Msg::App,Msg::Verbose);	
		double livetime = 0; 
		for(unsigned i=0; i<17; i++)
		{
			std::stringstream keyName;
  			keyName << "LIVETI";
  			keyName.width(2);
  			keyName.fill('0');
  			keyName << i;
  
			if( _eventTable->hasAttribute(keyName.str()) )			   			   		
	    		*_eventTable->addAttribute(keyName.str(), "Live time for a particular CCD") 
    	  		= livetime;	
  		}
		if( _eventTable->hasAttribute("LIVETIME") )			   			   		
	   		*_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") 
      		= livetime;	
	}


  return;
}


void XmmExposure::_getGtiLists() {

  // Initialize Gti lists with tstart and tstop of EXPOSUnn extensions

  for(unsigned i=0; i<_ccdNums.size(); i++){
    std::stringstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << _ccdNums[i];
 
    Block* expBlock = _set->block(extName.str());
    Gti expGti(expBlock->attribute("TSTART")->asReal(),
	       expBlock->attribute("TSTOP")->asReal());
 
    GtiList tempList;
    tempList.push_back(expGti);
    _gtiLists.push_back(tempList);
  }

  // Step thru each component of the DSS. Indetify which components have TIME filters and which
  // have CCDNR filters.  Store this information for use down stream.

  DataSubspace dss(_eventTable);
  unsigned ncomps = dss.components();
   
  if(ncomps>0){
    std::map<unsigned, DataSubspace *> ccdSubspaces;
    ccdSubspaces[0] = new DataSubspace;
    for(unsigned i=0; i<_ccdNums.size(); i++) ccdSubspaces[_ccdNums[i]] = new DataSubspace;
    
    for(unsigned icomp=0; icomp<ncomps; icomp++){
      bool noCcds = true;
      DScomp * comp = dss.component(icomp);
      for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++){
	DSfilter *filter = comp->filter(ifilt);
	DSS::DSfilterComp comps = filter->comps();

	if( filter->name()=="CCDNR" ){
	  if( filter->type()==DSS::RangeT ){
    	    std::list<Range<double> > ranges = filter->ranges();
    	    std::list<Range<double> >::iterator irange;
    	    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
    	      for(unsigned i=0; i<_ccdNums.size(); i++){
		if( irange->contains((double)_ccdNums[i]) )
    		  ccdSubspaces[_ccdNums[i]]->addComponent(*comp);
    	      }
    	    }
	    if( !noCcds ){
	      errstr << "The filtering component `" << comp->selectExpression() << "' contains more than "
		     << "one CCDNR filter.  As a result the exposure updating calculation may produce "
		     << "an incorrect result." << Sas::warning("ExpMultiCcdFilters");
	    }
    	    noCcds = false;
    	  }else{
    	    errstr << "The filtering expression `" << filter->selectExpression() << "' may not allow for "
    		   << "calculation of a uniform exposure for a CCD.  This filter is being ignored "
    		   << "in the exposure updating calculation."
    		   << Sas::warning("ExpBadCcdType");
    	  }
    	}else if( comps[0]=="CCDNR" || comps[1]=="CCDNR" ){
    	  errstr << "The filtering expression `" << filter->selectExpression() << "' does not allow for "
    		 << "calculation of a uniform exposure for a CCD.  This filter is being ignored "
    		 << "in the exposure updating calculation."
    		 << Sas::warning("ExpBadCcdComp");
    	}else if( filter->name()=="TIME" ){
    	  if( filter->type()!=DSS::RangeT && filter->type()!=DSS::GtiT){
    	    errstr << "The filtering expression `" << filter->selectExpression() << "' may not allow for "
    		   << "calculation of a uniform exposure for all events.  This filter is being ignored "
    		   << "in the exposure updating calculation."
    		   << Sas::warning("ExpBadTimeType");	    
    	  }
    	}else if( comps[0]=="TIME" || comps[1]=="TIME" ){
    	  errstr << "The filtering expression `" << filter->selectExpression() << "' does not allow for "
   		 << "calculation of a uniform exposure for a CCD.  This filter is being ignored "
    		 << "in the exposure updating calculation."
    		 << Sas::warning("ExpBadTimeComp");
    	}
      }
      
      if(noCcds) ccdSubspaces[0]->addComponent(*comp);
    }
    
    unsigned nccd=0;
    for(unsigned i=0; i<=_ccdNums.size(); i++){
      unsigned iccd;
      if( i==_ccdNums.size() ) iccd=0;
      else iccd = _ccdNums[i];
      DataSubspace * dss0 = ccdSubspaces[iccd];
      
      if( dss0->components() != 0 ){
	
    	// Need to check if TIME filters have been applied in a way that makes sense in terms
    	// of exposure correction, ie that different sets of events do not have different
    	// exposure corrections.
	
	// Do some special checks for global components
	if( iccd==0 && nccd!=0 ){
	  for(unsigned icomp=0; icomp<dss0->components(); icomp++){
	    for(unsigned ifilter=0; ifilter<dss0->component(icomp)->filters(); ifilter++){
	      if( dss0->component(icomp)->filter(ifilter)->name()!="TIME" ){
		errstr << "The filter expression `" << dss.selectExpression() 
		       << "' may not allow for uniform exposure correction. As a result the "
		       << "exposure updating calculation may produce an incorrect result."
		       << Sas::warning("ExpBadGlobalComp");
		break;
	      }
	    }
	  }
	}
	
	nccd++;

    	DataSubspace dssTemp = *dss0;
    	DScomp * comp0 = dssTemp.component(0);
        
    	// Check if we have just global TIME filters, or that all other filters are global.
    	
    	bool globalTime = true;
    	bool globalOther = true;
    	for(unsigned icomp=0; icomp<dss0->components(); icomp++){
   	  for(unsigned ifilter=0; ifilter<dss0->component(icomp)->filters(); ifilter++){
    	    DSfilter *filter = dss0->component(icomp)->filter(ifilter);
    	    if( !dssTemp.isGlobal(*filter) ){
    	      if(filter->name() == "TIME") globalTime = false;
    	      else globalOther = false;
    	    }
    	  }
    	}
	
    	if(globalTime){
    	  while( comp0->hasFilter("TIME", DSS::GtiT) ){
    	    DSfilter * filter0 = comp0->filter("TIME", DSS::GtiT);
    	    GtiList glist;
    	    InOGIPgtiTable oGtiTab(filter0->blockSpec());
    	    oGtiTab >> glist;
    	    if(iccd==0) for(unsigned j=0; j<_ccdNums.size(); j++) _gtiLists[j] &= glist;
    	    else _gtiLists[i] &= glist;
	    
    	    for(unsigned ifilter=0; ifilter<comp0->filters(); ifilter++)
    	      if( comp0->filter(ifilter) == filter0 ){
    		comp0->deleteFilter(ifilter);
   		break;
    	      }
    	  }
	  
    	  while( comp0->hasFilter("TIME", DSS::RangeT) ){
    	    DSfilter * filter0 = comp0->filter("TIME", DSS::RangeT);
    	    GtiList glist;
    	    std::list<Range<double> > ranges = filter0->ranges();
    	    std::list<Range<double> >::const_iterator irange;
    	    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
    	      double lower, upper;
    	      RangeBound lowerB, upperB;
    	      irange->values(lower, upper, lowerB, upperB);
    	      if( lowerB==Unbounded ) lower=0;
    	      if( upperB==Unbounded ) upper=Gti::ETERNITY;
    	      Gti gti(lower, upper);
    	      glist.push_back(gti);
    	    }
	    glist.sortIt();
    	    if(iccd==0) for(unsigned j=0; j<_ccdNums.size(); j++) _gtiLists[j] &= glist;
    	    else _gtiLists[i] &= glist;
    	  
    	    for(unsigned ifilter=0; ifilter<comp0->filters(); ifilter++)
    	      if( comp0->filter(ifilter) == filter0 ){
    		comp0->deleteFilter(ifilter);
    		break;
    	      }
    	  }
	  
    	}else{
    	  // TIME filters are not all global, so check if all other filters are global.
    	  // If not, warn the user that he is on his own.
    	  if( !globalOther )
   	    errstr << "A filtering expression has been specified which may not allow for "
    		   << "calculation of a uniform exposure for all events.  Please ensure that "
    		   << "all TIME related filters are applied uniformly."
   		   << Sas::warning("ExpBadDssGlobal");
	  
   	  GtiList glist;
    	  for(unsigned icomp=0; icomp<dssTemp.components(); icomp++){
    	    GtiList compGlist(true);
    	    DScomp * comp = dssTemp.component(icomp);
    	    for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
    	      DSfilter *filter = comp->filter(ifilter);
    	      if(filter->name()=="TIME"){
    		GtiList filterGlist;
    		if(filter->type()==DSS::GtiT){
    		  InOGIPgtiTable oGtiTab(filter->blockSpec());
    		  oGtiTab >> filterGlist;
    		  comp->deleteFilter(ifilter);
    		}else if(filter->type()==DSS::RangeT){
    		  std::list<Range<double> > ranges = filter->ranges();
    		  std::list<Range<double> >::const_iterator irange;
    		  for(irange=ranges.begin(); irange!=ranges.end(); irange++){
    		    double lower, upper;
    		    RangeBound lowerB, upperB;
    		    irange->values(lower, upper, lowerB, upperB);
    		    if( lowerB==Unbounded ) lower=0;
    		    if( upperB==Unbounded ) upper=Gti::ETERNITY;
   		    Gti gti(lower, upper);
    		    filterGlist.push_back(gti);
    		  }
    		  comp->deleteFilter(ifilter);
    		}
    		compGlist &= filterGlist;
    	      }
    	    }
    	    glist |= compGlist;
    	  }
    	  if(iccd==0) for(unsigned j=0; j<_ccdNums.size(); j++) _gtiLists[j] &= glist;
    	  else _gtiLists[i] &= glist;
    	}
      }
    }
    for(unsigned i=0; i<_ccdNums.size(); i++) delete ccdSubspaces[_ccdNums[i]];
    delete ccdSubspaces[0];
  }//end if(ncomps>0) 
  
  for(unsigned i=0; i<_ccdNums.size(); i++){
      _gtiLists[i].reduce();
      _gtiLists[i].sortIt();
  }

  // If requested, filter the EXPOSURE extensions with their corresponding GTI.
  

  if( booleanParameter("keepfilteroutput") && booleanParameter("filterexposure") ){
    std::string tempSetName = tempnam(0,0);
    for(unsigned i=0; i<_ccdNums.size(); i++){
      std::stringstream extName;
      extName << "EXPOSU";
      extName.width(2);
      extName.fill('0');
      extName << _ccdNums[i];

      if( _gtiLists[i].size() == 0 ){
		Table * expTab = _set->table(extName.str());
		expTab->deleteRows(0, expTab->rows());
      }else{
		DataSet *tempSet = dataSetServer->open(tempSetName, DataSet::Create);
		OutOGIPgtiTable tempgti(tempSet);
		tempgti << _gtiLists[i];
		dataSetServer->close(tempSet);


	/*
	 * AI: SPR-2632. If there is a PHASE keyword in the expression, we have also to 
	 * filtering the EXPOSUnn table with the PHASE expression. So, we add it to the final
	 * expression. Right now we only allow one PHASE expression
	 * TBD: Allow the user the possibility to use more than one PHASE expression.
	 */

	std::vector<std::string> expr;

	for(unsigned icomp=0; icomp<ncomps; icomp++)
	  {	  
	    DScomp * comp = dss.component(icomp);
	    
	    for(unsigned ifilt=0; ifilt<comp->filters(); ifilt++)
	      {
		DSfilter *filter = comp->filter(ifilt);
		
		if (filter->selectExpression().find("PHASE") != std::string::npos)
		  expr.push_back(filter->selectExpression());
	      }
	  }
	
	
	if (expr.size() == 1)
	  {
	    DataFilter filter( "gti(" + tempSetName + ",TIME) &&" + expr[0] );
	    filter.doFilter(*_set->table(extName.str()));
	  }
	else
	  {
	    if(expr.size() > 1 )
	      {
		errstr << "PHASE expression to complex. Please check the LIVETIME keyword value"
		       << Sas::warning("PhaseExpressionToComplex"); 		
	      }
	    
	    DataFilter filter( "gti(" + tempSetName + ",TIME)");
	    filter.doFilter(*_set->table(extName.str()));
	  }

	unlink(tempSetName.c_str());
      }
    }
  }

  return;
}


real64 XmmExposure::_getLiveTime(unsigned ccdRef){

    unsigned ccdNr = _ccdNums[ccdRef];

    std::stringstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << ccdNr;

    Table * expTab = _set->table(extName.str());

    real32 delTime = 0.;
    bool hasTimeDelCol = true;
    if( !expTab->hasColumn("TIMEDEL") ){
      hasTimeDelCol = false;
      delTime = expTab->attribute("TIMEDEL")->asReal();
    }

    ColumnData * timeCdat = expTab->column("TIME")->data();
    ColumnData * fracExpCdat = expTab->column("FRACEXP")->data();
    ColumnData * delTimeCdat = 0;
    if(hasTimeDelCol) delTimeCdat = expTab->column("TIMEDEL")->data();
    
    real64 * timeDat = timeCdat->real64Data();
    real32 * fracExpDat = fracExpCdat->real32Data();
    real32 * delTimeDat = 0;
    if(hasTimeDelCol) delTimeDat = delTimeCdat->real32Data();

    real64 timePixR = 0.5;
    if( expTab->hasAttribute("TIMEPIXR") ) 
      timePixR = expTab->attribute("TIMEPIXR")->asReal();

    GtiList gtilist = _gtiLists[ccdRef];
    const Gti * gti;

    real64 livetime = 0;
    unsigned long mRows = expTab->rows();
    for(unsigned long irow=0; irow<mRows; irow++){
      if(hasTimeDelCol) delTime = delTimeDat[irow];
      if( (gti=gtilist.inside(timeDat[irow]-delTime*timePixR)) != NULL && 
	  gti->stop() > timeDat[irow] + delTime*timePixR )
	livetime += delTime*fracExpDat[irow];
    }

    delete timeCdat;
    if(hasTimeDelCol) delete delTimeCdat;
    delete fracExpCdat;

    return livetime;
}


real64 XmmExposure::liveTime(unsigned ccdNr){
  unsigned iccdRef;
  for(iccdRef=0; iccdRef<_ccdNums.size(); iccdRef++)
    if( ccdNr == _ccdNums[iccdRef] ) break;

  if( iccdRef == _ccdNums.size() ) return 0;

  return _getLiveTime(iccdRef);
}

// writeDSlineKeys: finds the mean and max number of discarded lines
// during the extraction times and writes into header keywords
void XmmExposure::writeDSlineKeys() {

  vector<int> dslvec;

  // Get GTI lists if not already set
  if (_gtiLists.size() == 0)
    _getGtiLists();

  // Make a vector of dsline numbers from each CCD used in extraction
  for(unsigned iccdRef=0; iccdRef<_ccdNums.size(); iccdRef++)
     getDSlines(iccdRef, dslvec);

  // Write the discarded line keywords if numbers have been found
  if (dslvec.size()>0) {

     int dslmax = *std::max_element(dslvec.begin(), dslvec.end());
     double dslmean = std::accumulate(dslvec.begin(), dslvec.end(),0.0) 
                                   / dslvec.size();

     errstr << "Discarded lines: MAX: "<<dslmax<<" MEAN: "<<dslmean
	 << Sas::message(Msg::App,Msg::Verbose);	

     *_set->addAttribute("DLMAX", "Maximum number of discarded lines")      = dslmax;

     *_set->addAttribute("DLMEAN", "Mean number of discarded lines")       = dslmean;

  } else {
   // errstr << "No discarded lines found" << Sas::warning("NoDiscLines");
  }
}

// getDSlines: writes discarded line records into vector
void XmmExposure::getDSlines(unsigned iccdRef, vector<int> &dslvec) {

   // Attempt to open the housekeeping block
  std::stringstream hkName;
  hkName.width(2); hkName.fill('0'); hkName << _ccdNums[iccdRef];
  string hkname = "HKAUX" + hkName.str();

  if (_set->hasBlock(hkname)) {
     Table * hkTab = _set->table(hkname);

     // Map the discarded line values
     vector<double> time;
     vector<int> dslin;
     CalOalUtility::fillVector(hkTab, "TIME", time);
     CalOalUtility::fillVector(hkTab, "DSLIN", dslin);

     // Add records which fall within extraction times
     GtiList gtilist = _gtiLists[iccdRef];
     for (unsigned i=0; i<hkTab->rows(); ++i) 
         if (gtilist.inside(time[i]))
            dslvec.push_back(dslin[i]);
  }
}

//static AttString findNextAttribute(const std::string &s, unsigned i0)
//{
  // A bit HeathRobinsonesque but here we will look for the strings:
  // IMAGE, "RA,DEC"
  //
  
        //struct AttString a = { std::string::npos, 0 };
//        cout<<"S: "<<s<<endl;
//        AttString a = { std::string::npos, 0 };
//        unsigned l = s.length(), i = i0;
  //      while (i < l)
    //            if (s[i++] == '#') {
                        //
                        //      begining of attribute reference
                        //
      //                  a.pos = i-1;
        //                unsigned j = i;
          //              while (j<l && (isalnum(s[j]) || s[j] == '_' || s[j] == '-'))
            //                    ++j;
              //          a.len = j - a.pos;
                //        break;
//                }
  //      cout<<"Ain: "<<a.pos<<" "<<a.len<<endl;
    //    return a;
//} // findNextAttribute()

//struct AttString { std::string::size_type pos; unsigned len; };

vector<XmmExposure::AttString> XmmExposure::findSpecialStrings(std::string &s)
{
  // A bit HeathRobinsonesque but here we will look for the strings:
  // IMAGE, "RA,DEC"
  //
  vector<XmmExposure::AttString >a;
  // Find all occurrences of "IMAGE"
  string su = uppercase(s);
  string stest = "IMAGE";
  vector<size_t> pos = findAllOccurances(su,stest);
  // Loop over each occurrence of "IMAGE(fname,unit,unit,val)"
  for (unsigned i=0; i<pos.size(); i++) {
      // Check if this is part of a contiguous construct "IMAGE("
      string dummy = su.substr(pos[i]);
         // Remove whitespace
      dummy.erase(remove_if(dummy.begin(), dummy.end(),
                                      ::isspace),dummy.end());
      if (dummy.substr(0,6) == "IMAGE(") {

         // Search for closing bracket
         size_t end = s.find(")",pos[i]);
         if (end != string::npos) {
            XmmExposure::AttString ai = {pos[i],unsigned(end-pos[i]+1)}; 
            a.push_back(ai);
         } else {
            errstr<<"Incomplete expression: "<<s
            << Sas::error("IncompletExp");
         }
      }
  }

  // Find all occurrences of "RA,DEC"
  stest = "RA";
  pos = findAllOccurances(su,stest);
  // Loop over each "(RA,DEC) in <shape>(coord,coord,radius,raius,rot)"
  for (unsigned i=0; i<pos.size(); i++) {
      // Check if this is part of a contiguous construct "RA,DEC"
      string dummy = su.substr(pos[i]);
         // Remove whitespace
      dummy.erase(remove_if(dummy.begin(), dummy.end(), 
                                      ::isspace),dummy.end());
      if (dummy.substr(0,6) == "RA,DEC") {
         // Search for previous opening bracket and second closing
         size_t start= s.rfind("(",pos[i]);
         size_t end1 = s.find(")",pos[i]);
         size_t end = s.find(")",end1+1);
         size_t len = end-start+1;
         if (start != string::npos && end != string::npos) {
            XmmExposure::AttString ai = {start,len};
            a.push_back(ai);
         } else {
            errstr << "Incomplete expression: "<<s
            << Sas::error("IncompletExp");
         }
      }
  }

  return a;
}

/*
 *Find all positions of the a SubString in given String
 */
vector<size_t> XmmExposure::findAllOccurances(std::string &s, std::string &sub)
{
        vector<size_t> vec;

	// Get the first occurrence
	size_t pos = s.find(sub);
 
	// Repeat till end is reached
	while( pos != std::string::npos)
	{
		// Add position to the vector
		vec.push_back(pos);
 
		// Get the next occurrence from the current position
		pos = s.find(sub, pos + sub.size());
	}
        return vec;
}
 

// Method to convert the IMAGE construct into a BOX && MASK
// Needed because MASK is an exclusive construct
string XmmExposure::convSpecialExpression(string &e, string &fname)
{

    string expcopy = e;

  // Find the position of the special strings that need to be changed
    vector<XmmExposure::AttString> a = findSpecialStrings(expcopy);

  // Loop over these strings
    for (unsigned i=0; i<a.size(); i++) {

        // Check if the attribute is IMAGE
        std::string att(expcopy.substr(a[i].pos, a[i].len));
        if (uppercase(att).find("IMAGE") != string::npos) // or something
         {
           string newbit = interpretImageExpr(att, fname);
           expcopy.replace(a[i].pos, a[i].len, newbit);
         }

        // Check if celestial coords 
        if (uppercase(att).find("DEC") != string::npos) 
         {
           string newbit = convExpressionCeltoSky(att, fname);
           expcopy.replace(a[i].pos, a[i].len, newbit);
         }
    }
    return expcopy;
}

// Method to interpret IMAGE as a self-contained box
string XmmExposure::interpretImageExpr(string &att, string &fname)
{
  // Format is: IMAGE(filename,<maskvalue=0>) 
  //            where filename is the name of an image in sky or detector coords
  //            and maskvalue is the value to use as bad area - defaults to 0
  // Deal in Ra,DEC as much as possible to allow non-XMM images
  // to be used.
  //
    size_t commapos = att.find(","),
           start = att.find("(")+1,
           end = att.find(")");

    // Find if there is a maskvalue specified
    // maskvalue not currently used
    //string maskvalue = "0";
    // We tolerate and ignore everything after a comma
    if (commapos != string::npos) 
       end = commapos;

    // Get image name
    string s = att.substr(start,(end-start));
    s.erase(remove( s.begin(), s.end(), '\"' ),s.end());
    s.erase(remove( s.begin(), s.end(), '\'' ),s.end());
    string imagename = s;
    //cout<<"Image: "<<imagename<<" "<<fname<<endl;

    // Set the coordinate convertor for the image
    SscWcsConvertor::setWcs(imagename);
    //SscWcsConvertor::setWcsRef(imagename);

    // Find num pixels in each dim
    Array *a;
    DataSet *iset = dataSetServer->open(imagename, DataSet::Read);
    if (imagename.find(":") != string::npos) 
        a = iset->array(imagename);
    else 
        a = iset->array(imagename+":");
    
    vector<unsigned long> npix = a->size();

    // Find the coordinate system of the image
    string ctype1,Xaxis, Yaxis;
    if (iset->hasAttribute("CTYPE1"))
       ctype1=iset->attribute("CTYPE1")->asString();
    else if (a->hasAttribute("CTYPE1"))
       ctype1=a->attribute("CTYPE1")->asString();
    else
       errstr << "Image file does not contain a CTYPE1 "
              << "keyword." <<  Sas::error("BadImage");

    if (ctype1.find("DET") != string::npos)
       Xaxis="DETX", Yaxis="DETY";
    else if (ctype1.find("RAW") != string::npos)
       errstr << "RAW pixel image not supported in the IMAGE construct"
              << Sas::error("NoRawPixImage");
    else // assume this is a sky or RA/DEC system 
       Xaxis="X", Yaxis="Y";

    CoordPair cel_bl,cel_tr;  // bottom-left and top-right RA/DEC coords
    CoordPair out_bl,out_tr;  // bottom-left and top-right output  coords

    // Get RA,DEC of the corners of the image
    cel_bl = SscWcsConvertor::imagePixelsToRaDec(1.0,1.0);
    cel_tr = SscWcsConvertor::imagePixelsToRaDec(double(npix[1]),
                                                 double(npix[0]));

    //cout<<"BL: "<<cel_bl.x()<<" "<<cel_bl.y()<<endl;
    //cout<<"TR: "<<cel_tr.x()<<" "<<cel_tr.y()<<endl;
    // Find the equivalent detector or sky coordinates for the event file
    // Change to the event file WCS
    SscWcsConvertor::setWcs(fname);
    SscWcsConvertor::setWcsRef(fname);
    
    if (Xaxis=="DETX") {
       out_bl = SscWcsConvertor::raDecToDet(cel_bl.x(), cel_bl.y());
       out_tr = SscWcsConvertor::raDecToDet(cel_tr.x(), cel_tr.y());
    } else { // Sky coords
       out_bl = SscWcsConvertor::raDecToSkyPixels(cel_bl.x(), cel_bl.y());
       out_tr = SscWcsConvertor::raDecToSkyPixels(cel_tr.x(), cel_tr.y());
    }
    //cout<<"OBL: "<<out_bl.x()<<" "<<out_bl.y()<<endl;
    //cout<<"OTR: "<<out_tr.x()<<" "<<out_tr.y()<<endl;

    // Find the half-width
    double xhwid = (out_tr.x() - out_bl.x()) /2.0;
    double yhwid = (out_tr.y() - out_bl.y()) /2.0;
    double xcen = (out_tr.x() + out_bl.x()) /2.0;
    double ycen = (out_tr.y() + out_bl.y()) /2.0;

    // Define a BOX expression encompassing the Image
    string b = "("+toString(Xaxis)+","+toString(Yaxis)+") in BOX("+
               toString(xcen)+","+toString(ycen)+","+toString(xhwid)
                  +","+toString(yhwid)+",0)";

    // Add the original IMAGE expression as a MASK - but keep image name 
    // in original case
    string mask = "MASK("+ imagename + ",0,0," + Xaxis + 
                     "," + Yaxis + ")";
                     //"," + Yaxis + "," + maskvalue + ")";
    //string c = att;
    //c.replace(uppercase(c).find("IMAGE"),5,"MASK");
    b = b + "&&" + mask;

    return b;
}

// Method to convert expression from RA/DEC to X/Y coords
string XmmExposure::convExpressionCeltoSky(string &att, string &fname)
{
  // Find the coordinates and radius and convert to POS frame
  vector<double> vals = extractVals(att);

  //cout<<vals[0]<<""<<vals[1]<<""<<vals[2]<<endl;
  // Convert centre from RA,DEC to X/Y
  SscWcsConvertor::setWcs(fname);
  SscWcsConvertor::setWcsRef(fname);
  CoordPair cel(vals[0] * DEG_TO_RAD,vals[1] * DEG_TO_RAD);
  CoordPair c = SscWcsConvertor::raDecToSkyPixels(cel.x(), cel.y());

  // Make the output string
  string b;
  if (uppercase(att).find("BOX") != string::npos)
     b = "(X,Y) in BOX(";
  else if (uppercase(att).find("CIRCLE") != string::npos) 
     b="(X,Y) in CIRCLE(";
  else if (uppercase(att).find("ANNULUS") != string::npos) 
     b="(X,Y) in ANNULUS(";
  else
     errstr << "Can not convert the expression "<<att
            << Sas::error("CelExpNotRec");

  // Add the centre coords in X/Y
  b = b + toString(c.x()) + "," + toString(c.y());

  // Add radii or half-widths - what about rotation ?
  for (unsigned i=2; i<vals.size(); i++) 
     b = b + "," + toString(vals[i] * 72000.0);

  b = b +  ")";

  return b;
}
// Method to extract numbers from a spatial expression
vector<double> XmmExposure::extractVals(string &att)
{
  vector<double> vals;

  // Extract the RA, DEC centre
  size_t p1 = uppercase(att).find("DEC");
  if (p1 != string::npos) {
     // Find opening brackets
     size_t pstart = att.find("(",p1) + 1;
     size_t pend = att.find(")",pstart);

     // Find all commas
     size_t pos = att.find(",",pstart);

     while (pos != string::npos)
     {
        // Save the value
        vals.push_back(std::stod(att.substr(pstart,pos)));
        pstart = pos+1;
        pos = att.find(",",pstart);
     }
     vals.push_back(std::stod(att.substr(pstart,pend)));
  }
  else 
     errstr << "RA/DEC not found in string to convert"
            << Sas::error("CelExpNotRec");

  return vals;
}


